@@ -66,6 +66,7 @@ from nuitka.utils.Utils import (
     isLinux,
     isMacOS,
     isOpenBSD,
+    isWin32OrPosixWindows,
     isWin32Windows,
 )
 
@@ -101,6 +102,21 @@ def _checkSpec(value, arg_name):
             % (arg_name, value)
         )
 
+    if value.count("%") % 2 != 0:
+        Tracing.options_logger.warning(
+            """Unmatched '%%' is suspicious for '%s=%s' and may \
+not do what you want it to do."""
+            % (arg_name, value)
+        )
+
+    for candidate in ("%PROGRAM%", "%CACHE_DIR%", "%HOME%", "%TEMP%"):
+        if candidate in value[1:]:
+            Tracing.options_logger.sysexit(
+                """Absolute runtime path of '%s' can only be at the \
+start of '%s=%s', using it in the middle is not allowed."""
+                % (candidate, arg_name, value)
+            )
+
 
 def _checkOnefileTargetSpec():
     _checkSpec(options.onefile_tempdir_spec, arg_name="--onefile-tempdir-spec")
@@ -113,13 +129,6 @@ you cannot unpack the onefile payload into the same directory as the binary,
 as that would overwrite it and cause locking issues as well."""
         )
 
-    if options.onefile_tempdir_spec.count("%") % 2 != 0:
-        Tracing.options_logger.warning(
-            """Unmatched '%%' is suspicious for '--onefile-tempdir-spec' and may \
-not do what you want it to do: '%s'"""
-            % options.onefile_tempdir_spec
-        )
-
     if options.onefile_tempdir_spec.count("%") == 0:
         Tracing.options_logger.warning(
             """Not using any variables for '--onefile-tempdir-spec' should only be \
@@ -143,6 +152,20 @@ very well known environment, anchor with e.g. %%TEMP%%, %%CACHE_DIR%% is recomme
         )
 
 
+def _checkAutoUpdateUrlSpec():
+    _checkSpec(options.auto_update_url_spec, arg_name="--auto-update-url-spec")
+
+    if not options.auto_update_url_spec.startswith(
+        (
+            "http://",
+            "https://",
+        )
+    ):
+        Tracing.optimization_logger.sysexit(
+            "Error, only 'https://' and 'http://' URLs are allowed."
+        )
+
+
 def parseArgs():
     """Parse the command line arguments
 
@@ -180,10 +203,10 @@ def parseArgs():
     if options.quiet or int(os.environ.get("NUITKA_QUIET", "0")):
         Tracing.setQuiet()
 
-    if not shallDumpBuiltTreeXML():
-        Tracing.options_logger.info(
-            "Used command line options: %s" % " ".join(sys.argv[1:])
-        )
+    Tracing.options_logger.info(
+        "Used command line options: %s"
+        % " ".join(('"%s"' % arg) if " " in arg else arg for arg in sys.argv[1:])
+    )
 
     if os.environ.get("NUITKA_REEXECUTION") and not isAllowedToReexecute():
         Tracing.general.sysexit(
@@ -227,6 +250,10 @@ def parseArgs():
     if options.onefile_tempdir_spec:
         _checkOnefileTargetSpec()
 
+    # Check auto update URL spec
+    if options.auto_update_url_spec:
+        _checkAutoUpdateUrlSpec()
+
     # Provide a tempdir spec implies onefile tempdir, even on Linux.
     # Standalone mode implies an executable, not importing "site" module, which is
     # only for this machine, recursing to all modules, and even including the
@@ -284,15 +311,29 @@ Error, '--nofollow-import-to' takes only module names or patterns, not directory
             "Error, no such Python binary %r, should be full path." % scons_python
         )
 
-    if options.output_filename is not None and (
-        (isStandaloneMode() and not isOnefileMode()) or shallMakeModule()
-    ):
-        Tracing.options_logger.sysexit(
-            """\
+    output_filename = getOutputFilename()
+
+    if output_filename is not None:
+        if (isStandaloneMode() and not isOnefileMode()) or shallMakeModule():
+            Tracing.options_logger.sysexit(
+                """\
 Error, may only specify output filename for acceleration and onefile mode,
 but not for module mode where filenames are mandatory, and not for
 standalone where there is a sane default used inside the dist folder."""
-        )
+            )
+
+        output_dir = os.path.dirname(output_filename) or "."
+
+        if not os.path.isdir(output_dir):
+            Tracing.options_logger.sysexit(
+                """\
+Error, specified output directory does not exist, you have to create
+it before using it: '%s' (from --output-filename='%s')."""
+                % (
+                    output_dir,
+                    output_filename,
+                )
+            )
 
     if isLinux():
         if len(getIconPaths()) > 1:
@@ -332,6 +373,11 @@ standalone where there is a sane default used inside the dist folder."""
             "Error, icon path %r does not exist." % icon_exe_path
         )
 
+    if isMacOS() and not shallCreateAppBundle() and shallDisableConsoleWindow():
+        Tracing.options_logger.sysexit(
+            "Error, cannot disable console unless also using '--macos-create-app-bundle'."
+        )
+
     try:
         file_version = getWindowsFileVersion()
     except Exception:  # Catch all the things, don't want any interface, pylint: disable=broad-except
@@ -379,7 +425,7 @@ standalone where there is a sane default used inside the dist folder."""
             "Conflicting options '--follow-imports' and '--nofollow-imports' given."
         )
 
-    for module_pattern in getShallIncludePackageData():
+    for module_pattern, _filename_pattern in getShallIncludePackageData():
         if (
             module_pattern.startswith("-")
             or "/" in module_pattern
@@ -554,7 +600,7 @@ def commentArgs():
         )
 
     # TODO: Not all of these are usable with MSYS2 really, split those off.
-    if getOS() != "Windows":
+    if not isWin32OrPosixWindows():
         # Too many Windows specific options clearly
         if (
             getWindowsIconExecutablePath()
@@ -628,8 +674,7 @@ to work. You can selectively add them with '--follow-import-to=name' though."""
         )
 
     if (
-        not shallDumpBuiltTreeXML()
-        and not standalone_mode
+        not standalone_mode
         and not options.follow_all
         and not options.follow_none
         and not options.follow_modules
@@ -680,7 +725,6 @@ make sure that is intended."""
     if (
         options.static_libpython == "auto"
         and not shallMakeModule()
-        and not shallDumpBuiltTreeXML()
         and not shallUseStaticLibPython()
         and getSystemStaticLibPythonPath() is not None
     ):
@@ -721,6 +765,18 @@ and not with the non-debug version.
 """
         )
 
+    if (
+        isMacOS()
+        and shallCreateAppBundle()
+        and shallDisableConsoleWindow()
+        and not getIconPaths()
+    ):
+        Tracing.general.warning(
+            """\
+For GUI applications, you ought to specify an icon with '--macos-app-icon'.", \
+otherwise a dock icon may not be present."""
+        )
+
     filename = getPositionalArgs()[0]
     if not os.path.exists(filename):
         Tracing.general.sysexit("Error, file '%s' is not found." % filename)
@@ -754,9 +810,9 @@ def shallRunInDebugger():
     return options.debugger
 
 
-def shallDumpBuiltTreeXML():
-    """:returns: bool derived from ``--xml``"""
-    return options.dump_xml
+def getXMLDumpOutputFilename():
+    """:returns: str derived from ``--xml``"""
+    return options.xml_output
 
 
 def shallOnlyExecCCompilerCall():
@@ -863,8 +919,22 @@ def getMustIncludePackages():
 
 
 def getShallIncludePackageData():
-    """*list*, items of ``--include-package-data=``"""
-    return sum([_splitShellPattern(x) for x in options.package_data], [])
+    """*iterable of (module pattern, filename pattern)*, derived from ``--include-package-data=``
+
+    The filename pattern can be None if not given. Empty values give None too.
+    """
+    for package_data_pattern in sum(
+        [_splitShellPattern(x) for x in options.package_data], []
+    ):
+        if ":" in package_data_pattern:
+            module_pattern, filename_pattern = package_data_pattern.split(":", 1)
+            # Empty equals None.
+            filename_pattern = filename_pattern or None
+        else:
+            module_pattern = package_data_pattern
+            filename_pattern = None
+
+        yield module_pattern, filename_pattern
 
 
 def getShallIncludeDataFiles():
@@ -934,17 +1004,17 @@ def shallUsePythonDebug():
     return options.python_debug or sys.flags.debug
 
 
-def isUnstriped():
-    """:returns: bool derived from ``--unstriped`` or ``--profile``
+def isUnstripped():
+    """:returns: bool derived from ``--unstripped`` or ``--profile``
 
     A binary is called stripped when debug information is not present, an
-    unstriped when it is present. For profiling and debugging it will be
+    unstripped when it is present. For profiling and debugging it will be
     necessary, but it doesn't enable debug checks like ``--debug`` does.
 
-    Passed to Scons as ``unstriped_mode`` to it can ask the linker to
+    Passed to Scons as ``unstripped_mode`` to it can ask the linker to
     include symbol information.
     """
-    return options.unstriped or options.profile
+    return options.unstripped or options.profile
 
 
 def isProfile():
@@ -1256,7 +1326,6 @@ def isOnefileTempDirMode():
     spec = getOnefileTempDirSpec()
 
     for candidate in (
-        "%TEMP%",
         "%PID",
         "%TIME%",
         "%PROGRAM%",
@@ -1309,11 +1378,17 @@ def getPythonPgoUnseenModulePolicy():
 
 
 def getOnefileTempDirSpec():
+    """*str* = ``--onefile-tempdir-spec``"""
     return (
         options.onefile_tempdir_spec or "%TEMP%" + os.path.sep + "onefile_%PID%_%TIME%"
     )
 
 
+def getAutoUpdateUrlSpec():
+    """*str* = ``--onefile-tempdir-spec``"""
+    return options.auto_update_url_spec
+
+
 def getIconPaths():
     """*list of str*, values of ``--windows-icon-from-ico`` and ``--linux-onefile-icon``"""
 
