@@ -1,4 +1,4 @@
-#     Copyright 2021, Kay Hayen, mailto:kay.hayen@gmail.com
+#     Copyright 2022, Kay Hayen, mailto:kay.hayen@gmail.com
 #
 #     Part of "Nuitka", an optimizing Python compiler that is compatible and
 #     integrates with CPython, but also works on its own.
@@ -26,25 +26,83 @@ The base class will serve as documentation. And it will point to examples of
 it being used.
 """
 
+import ast
+import functools
 import inspect
-import os
-import shutil
 import sys
 from collections import namedtuple
 
 from nuitka.__past__ import getMetaClassBase
-from nuitka.freezer.IncludedEntryPoints import makeDllEntryPoint
-from nuitka.Options import isStandaloneMode
+from nuitka.containers.OrderedSets import OrderedSet
+from nuitka.freezer.IncludedDataFiles import (
+    makeIncludedDataDirectory,
+    makeIncludedDataFile,
+    makeIncludedEmptyDirectory,
+    makeIncludedGeneratedDataFile,
+    makeIncludedPackageDataFiles,
+)
+from nuitka.freezer.IncludedEntryPoints import (
+    makeDllEntryPoint,
+    makeExeEntryPoint,
+)
+from nuitka.ModuleRegistry import (
+    addModuleInfluencingCondition,
+    getModuleInclusionInfoByName,
+)
+from nuitka.Options import isStandaloneMode, shallMakeModule
+from nuitka.PythonFlavors import isAnacondaPython, isDebianPackagePython
+from nuitka.PythonVersions import getSupportedPythonVersions, python_version
 from nuitka.Tracing import plugins_logger
 from nuitka.utils.Execution import NuitkaCalledProcessError, check_output
-from nuitka.utils.FileOperations import makePath
-from nuitka.utils.ModuleNames import ModuleName
+from nuitka.utils.ModuleNames import (
+    ModuleName,
+    makeTriggerModuleName,
+    post_module_load_trigger_name,
+    pre_module_load_trigger_name,
+)
 from nuitka.utils.SharedLibraries import locateDLL, locateDLLsInDirectory
+from nuitka.utils.Utils import isLinux, isMacOS, isWin32Windows
 
-pre_modules = {}
-post_modules = {}
+_warned_unused_plugins = set()
 
-warned_unused_plugins = set()
+# TODO: Could share data cache with meta data nodes
+_package_versions = {}
+
+
+def _convertVersionToTuple(version_str):
+    def numberize(v):
+        # For now, we ignore rc/post stuff, hoping it doesn't matter for us.
+        return int("".join(d for d in v if d.isdigit()))
+
+    return tuple(numberize(d) for d in version_str.split("."))
+
+
+def _getPackageVersion(distribution_name):
+    if distribution_name not in _package_versions:
+        try:
+            if python_version >= 0x380:
+                from importlib.metadata import version
+            else:
+                from importlib_metadata import version
+
+            result = _convertVersionToTuple(version(distribution_name))
+        except ImportError:
+            try:
+                from pkg_resources import get_distribution
+
+                result = _convertVersionToTuple(
+                    get_distribution(distribution_name).version
+                )
+            except ImportError:
+                # Fallback if nothing is available, which may happen if no package is installed,
+                # but only source code is found.
+                result = _convertVersionToTuple(
+                    __import__(distribution_name).__version__.split(".")
+                )
+
+        _package_versions[distribution_name] = result
+
+    return _package_versions[distribution_name]
 
 
 class NuitkaPluginBase(getMetaClassBase("Plugin")):
@@ -53,11 +111,11 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
     Derive your plugin from "NuitkaPluginBase" please.
     For instructions, see https://github.com/Nuitka/Nuitka/blob/orsiris/UserPlugin-Creation.rst
 
-    Plugins allow to adapt Nuitka's behaviour in a number of ways as explained
+    Plugins allow to adapt Nuitka's behavior in a number of ways as explained
     below at the individual methods.
 
     It is used to deal with special requirements some packages may have (e.g. PyQt
-    and tkinter), data files to be included (e.g. certifi), inserting hidden
+    and tkinter), data files to be included (e.g. "certifi"), inserting hidden
     code, coping with otherwise undetectable needs, or issuing messages in
     certain situations.
 
@@ -113,18 +171,23 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
             True or False
 
         """
-        return True
+        return not cls.isDeprecated()
+
+    @classmethod
+    def isDeprecated(cls):
+        """Is this a deprecated plugin, i.e. one that has no use anymore."""
+        return False
+
+    @classmethod
+    def isDetector(cls):
+        """Is this a detection plugin, i.e. one which is only there to inform."""
+        return hasattr(cls, "detector_for")
 
     @classmethod
     def addPluginCommandLineOptions(cls, group):
         # Call group.add_option() here.
         pass
 
-    @classmethod
-    def getTagDataFileTagOptions(cls):
-        # Return tag_name, description tuples
-        return ()
-
     @classmethod
     def getPluginDefaultOptionValues(cls):
         """This method is used to get a values to use as defaults.
@@ -260,6 +323,42 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
         # Virtual method, pylint: disable=unused-argument
         return None
 
+    @staticmethod
+    def createFakeModuleDependency(module):
+        """Create module to depend on.
+
+        Notes:
+            Called by @onModuleDiscovered.
+
+        Args:
+            module: the module object
+
+        Returns:
+            None (does not apply, default)
+            tuple (code, reason)
+            tuple (code, reason, flags)
+        """
+        # Virtual method, pylint: disable=unused-argument
+        return None
+
+    @staticmethod
+    def hasPreModuleLoadCode(module_name):
+        return (
+            getModuleInclusionInfoByName(
+                makeTriggerModuleName(module_name, pre_module_load_trigger_name)
+            )
+            is not None
+        )
+
+    @staticmethod
+    def hasPostModuleLoadCode(module_name):
+        return (
+            getModuleInclusionInfoByName(
+                makeTriggerModuleName(module_name, post_module_load_trigger_name)
+            )
+            is not None
+        )
+
     def onModuleDiscovered(self, module):
         """Called with a module to be loaded.
 
@@ -277,19 +376,44 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
         # Virtual method, pylint: disable=no-self-use,unused-argument
         return None
 
-    def onModuleEncounter(self, module_filename, module_name, module_kind):
+    def getPackageExtraScanPaths(self, package_name, package_dir):
+        """Provide other directories to consider submodules to live in.
+
+        Args:
+            module_name: full module name
+            package_dir: directory of the package
+
+        Returns:
+            Iterable list of directories, non-existent ones are ignored.
+        """
+
+        # Virtual method, pylint: disable=no-self-use,unused-argument
+        return ()
+
+    def onModuleEncounter(self, module_name, module_filename, module_kind):
         """Help decide whether to include a module.
 
         Args:
-            module_filename: filename
             module_name: full module name
-            module_kind: one of "py", "shlib" (shared library)
+            module_filename: filename
+            module_kind: one of "py", "extension" (shared library)
         Returns:
             True or False
         """
         # Virtual method, pylint: disable=no-self-use,unused-argument
         return None
 
+    def onModuleRecursion(self, module_name, module_filename, module_kind):
+        """React to recursion to a module coming up.
+
+        Args:
+            module_name: full module name
+            module_filename: filename
+            module_kind: one of "py", "extension" (shared library)
+        Returns:
+            None
+        """
+
     def onModuleInitialSet(self):
         """Provide extra modules to the initial root module set.
 
@@ -319,7 +443,6 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
         """Provide a filename / -path for a to-be-imported module.
 
         Args:
-            importing: module object that asked for it (tracing only)
             module_name: (str or ModuleName) full name of module
         Returns:
             filename for module
@@ -361,44 +484,37 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
         """
         return locateDLLsInDirectory(directory)
 
-    @classmethod
-    def makeDllEntryPoint(cls, source_path, dest_path, package_name):
+    def makeDllEntryPoint(self, source_path, dest_path, package_name, reason):
         """Create an entry point, as expected to be provided by getExtraDlls."""
         return makeDllEntryPoint(
-            source_path=source_path, dest_path=dest_path, package_name=package_name
+            logger=self,
+            source_path=source_path,
+            dest_path=dest_path,
+            package_name=package_name,
+            reason=reason,
+        )
+
+    def makeExeEntryPoint(self, source_path, dest_path, package_name, reason):
+        """Create an entry point, as expected to be provided by getExtraDlls."""
+        return makeExeEntryPoint(
+            logger=self,
+            source_path=source_path,
+            dest_path=dest_path,
+            package_name=package_name,
+            reason=reason,
         )
 
     def reportFileCount(self, module_name, count, section=None):
         if count:
-            msg = "Found %d %s DLLs from '%s' %sinstallation." % (
+            msg = "Found %d %s DLLs from %s%s installation." % (
                 count,
                 "file" if count < 2 else "files",
-                "" if not section else section,
+                "" if not section else (" '%s' " % section),
                 module_name.asString(),
             )
 
             self.info(msg)
 
-    def considerExtraDlls(self, dist_dir, module):
-        """Provide a tuple of names of binaries to be included.
-
-        Args:
-            dist_dir: the distribution folder
-            module: the module object needing the binaries
-        Returns:
-            tuple
-        """
-        # TODO: This should no longer be here, as this API is obsolete, pylint: disable=unused-argument
-        for included_entry_point in self.getExtraDlls(module):
-            # Copy to the dist directory, which normally should not be a plugin task, but is for now.
-            makePath(os.path.dirname(included_entry_point.dest_path))
-
-            shutil.copyfile(
-                included_entry_point.source_path, included_entry_point.dest_path
-            )
-
-            yield included_entry_point
-
     def getExtraDlls(self, module):
         """Provide IncludedEntryPoint named tuples describing extra needs of the module.
 
@@ -411,6 +527,21 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
         # Virtual method, pylint: disable=no-self-use,unused-argument
         return ()
 
+    def onCopiedDLL(self, dll_filename):
+        """Chance for a plugin to modify DLLs after copy, e.g. to compress it, remove attributes, etc.
+
+        Args:
+            dll_filename: the filename of the DLL
+
+        Notes:
+            Do not remove or add any files in this method, this will not work well, there
+            is e.g. getExtraDLLs API to add things. This is only for post processing as
+            described above.
+
+        """
+        # Virtual method, pylint: disable=no-self-use,unused-argument
+        return None
+
     def getModuleSpecificDllPaths(self, module_name):
         """Provide a list of directories, where DLLs should be searched for this package (or module).
 
@@ -447,6 +578,74 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
         # Virtual method, pylint: disable=no-self-use,unused-argument
         return ()
 
+    def makeIncludedDataFile(self, source_path, dest_path, reason, tags=""):
+        return makeIncludedDataFile(
+            source_path=source_path,
+            dest_path=dest_path,
+            reason=reason,
+            tracer=self,
+            tags=tags,
+        )
+
+    def makeIncludedGeneratedDataFile(self, data, dest_path, reason, tags=""):
+        return makeIncludedGeneratedDataFile(
+            data=data, dest_path=dest_path, reason=reason, tracer=self, tags=tags
+        )
+
+    def makeIncludedDataDirectory(
+        self,
+        source_path,
+        dest_path,
+        reason,
+        tags="",
+        ignore_dirs=(),
+        ignore_filenames=(),
+        ignore_suffixes=(),
+        only_suffixes=(),
+        normalize=True,
+    ):
+        return makeIncludedDataDirectory(
+            source_path=source_path,
+            dest_path=dest_path,
+            reason=reason,
+            tracer=self,
+            tags=tags,
+            ignore_dirs=ignore_dirs,
+            ignore_filenames=ignore_filenames,
+            ignore_suffixes=ignore_suffixes,
+            only_suffixes=only_suffixes,
+            normalize=normalize,
+        )
+
+    def makeIncludedEmptyDirectory(self, dest_path, reason, tags):
+        return makeIncludedEmptyDirectory(
+            dest_path=dest_path,
+            reason=reason,
+            tracer=self,
+            tags=tags,
+        )
+
+    def makeIncludedPackageDataFiles(
+        self, package_name, package_directory, pattern, reason, tags
+    ):
+        return makeIncludedPackageDataFiles(
+            tracer=self,
+            package_name=ModuleName(package_name),
+            package_directory=package_directory,
+            pattern=pattern,
+            reason=reason,
+            tags=tags,
+        )
+
+    def updateDataFileTags(self, included_datafile):
+        """Add or remove data file tags."""
+
+    def onDataFileTags(self, included_datafile):
+        """Action on data file tags."""
+
+    def onBeforeCodeParsing(self):
+        """Prepare for code parsing, normally not needed."""
+
     def onStandaloneDistributionFinished(self, dist_dir):
         """Called after successfully creating a standalone distribution.
 
@@ -483,6 +682,30 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
         # Virtual method, pylint: disable=no-self-use,unused-argument
         return None
 
+    def onBootstrapBinary(self, filename):
+        """Called after successfully creating a bootstrap binary, but without payload.
+
+        Args:
+            filename: the created bootstrap binary, will be modified later
+
+        Returns:
+            None
+        """
+        # Virtual method, pylint: disable=no-self-use,unused-argument
+        return None
+
+    def onStandaloneBinary(self, filename):
+        """Called after successfully creating a standalone binary.
+
+        Args:
+            filename: the created standalone binary
+
+        Returns:
+            None
+        """
+        # Virtual method, pylint: disable=no-self-use,unused-argument
+        return None
+
     def onFinalResult(self, filename):
         """Called after successfully finishing a compilation.
 
@@ -545,6 +768,16 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
         # Virtual method, pylint: disable=no-self-use
         return None
 
+    def getExtraIncludeDirectories(self):
+        """Decide which extra directories to use for C includes in compilation.
+
+        Returns:
+            List of directories or None by default
+        """
+
+        # Virtual method, pylint: disable=no-self-use
+        return None
+
     def getExtraCodeFiles(self):
         """Add extra code files to the compilation.
 
@@ -596,13 +829,22 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
         Returns:
             None
         """
-        if self.plugin_name not in warned_unused_plugins:
-            warned_unused_plugins.add(self.plugin_name)
+        if self.plugin_name not in _warned_unused_plugins:
+            _warned_unused_plugins.add(self.plugin_name)
 
             plugins_logger.warning(
                 "Use '--enable-plugin=%s' for: %s" % (self.plugin_name, message)
             )
 
+    def onDataComposerRun(self):
+        """Internal use only.
+
+        Returns:
+            None
+        """
+        # Virtual method, pylint: disable=no-self-use
+        return None
+
     def onDataComposerResult(self, blob_filename):
         """Internal use only.
 
@@ -624,7 +866,7 @@ class NuitkaPluginBase(getMetaClassBase("Plugin")):
     _runtime_information_cache = {}
 
     def queryRuntimeInformationMultiple(self, info_name, setup_codes, values):
-        info_name = self.plugin_name + "_" + info_name
+        info_name = self.plugin_name.replace("-", "_") + "_" + info_name
 
         if info_name in self._runtime_information_cache:
             return self._runtime_information_cache[info_name]
@@ -678,9 +920,8 @@ except ImportError:
 
         NamedTupleResult = namedtuple(info_name, keys)
 
-        # We are being lazy here, the code is trusted, pylint: disable=eval-used
         self._runtime_information_cache[info_name] = NamedTupleResult(
-            *(eval(value) for value in feedback)
+            *(ast.literal_eval(value) for value in feedback)
         )
 
         return self._runtime_information_cache[info_name]
@@ -692,9 +933,82 @@ except ImportError:
             values=(("key", value),),
         ).key
 
-    @staticmethod
-    def onFunctionBodyParsing(module_name, function_name, body):
-        pass
+    def onFunctionBodyParsing(self, module_name, function_name, body):
+        """Provide a different function body for the function of that module."""
+        # Virtual method, pylint: disable=no-self-use,unused-argument
+        return None
+
+    def getCacheContributionValues(self, module_name):
+        """Provide values that represent the include of a plugin on the compilation.
+
+        This must be used to invalidate cache results, e.g. when using the
+        onFunctionBodyParsing function, and other things, that do not directly
+        affect the source code.
+        """
+        # Virtual method, pylint: disable=no-self-use,unused-argument
+        return ()
+
+    def evaluateCondition(self, full_name, condition, control_tags=()):
+        # Note: Caching makes no sense yet, this should all be very fast and
+        # cache themselves. TODO: Allow plugins to contribute their own control
+        # tag values during creation and during certain actions.
+        if condition == "True":
+            return True
+        if condition == "False":
+            return False
+
+        context = TagContext(logger=self, full_name=full_name)
+        context.update(control_tags)
+
+        context.update(
+            {
+                "macos": isMacOS(),
+                "win32": isWin32Windows(),
+                "linux": isLinux(),
+                "anaconda": isAnacondaPython(),
+                "debian_python": isDebianPackagePython(),
+                "standalone": isStandaloneMode(),
+                "module_mode": shallMakeModule(),
+                # TODO: Allow to provide this.
+                "deployment": False,
+                # Module dependent
+                "main_module": full_name.getBasename() == "__main__",
+                # Querying package versions.
+                "version": _getPackageVersion,
+            }
+        )
+
+        versions = getSupportedPythonVersions()
+
+        for version in versions:
+            big, major = version.split(".")
+            numeric_version = int(big) * 256 + int(major) * 16
+            is_same_or_higher_version = python_version >= numeric_version
+
+            context["python" + big + major + "_or_higher"] = is_same_or_higher_version
+            context["before_python" + big + major] = not is_same_or_higher_version
+
+        context["before_python3"] = python_version < 0x300
+        context["python3_or_higher"] = python_version >= 0x300
+
+        # We trust the yaml files, pylint: disable=eval-used
+        result = eval(condition, context)
+
+        if type(result) is not bool:
+            self.sysexit(
+                "Error, condition '%s' for module '%s' did not evaluate to boolean result."
+                % (condition, full_name)
+            )
+
+        addModuleInfluencingCondition(
+            module_name=full_name,
+            plugin_name=self.plugin_name,
+            condition=condition,
+            control_tags=context.used_tags,
+            result=result,
+        )
+
+        return result
 
     @classmethod
     def warning(cls, message):
@@ -709,33 +1023,6 @@ except ImportError:
         plugins_logger.sysexit(cls.plugin_name + ": " + message)
 
 
-def isTriggerModule(module):
-    return module in pre_modules.values() or module in post_modules.values()
-
-
-def replaceTriggerModule(old, new):
-    found = None
-    for key, value in pre_modules.items():
-        if value is old:
-            found = key
-            break
-
-    if found is not None:
-        pre_modules[found] = new
-
-    found = None
-    for key, value in post_modules.items():
-        if value is old:
-            found = key
-            break
-
-    if found is not None:
-        post_modules[found] = new
-
-
-import functools
-
-
 def standalone_only(func):
     """For plugins that have functionality that should be done in standalone mode only."""
 
@@ -750,3 +1037,27 @@ def standalone_only(func):
                 return None
 
     return wrapped
+
+
+class TagContext(dict):
+    def __init__(self, logger, full_name, *args, **kwargs):
+        dict.__init__(self, *args, **kwargs)
+
+        self.logger = logger
+        self.full_name = full_name
+
+        self.used_tags = OrderedSet()
+
+    def __getitem__(self, key):
+        try:
+            self.used_tags.add(key)
+
+            return dict.__getitem__(self, key)
+        except KeyError:
+            if key.startswith("use_"):
+                return False
+
+            self.logger.sysexit(
+                "Identifier '%s' in 'when' configuration of module '%s' is unknown."
+                % (key, self.full_name)
+            )
